searchData={"content_type":"text/markdown","items":[{"doc":"Defines an Ecto query builder.\n\nIt accepts the following options:\n\n  * `:schema` - the `Ecto.Schema` for which the queries will be built (required)\n  * `:searchable_fields` - a list of fields to build default filters for. This option is simply a\n    shorthand syntax for using the `QueryElf.Plugins.AutomaticFilters` plugin with the given list\n    as the `fields` option. You should check the plugin's documentation for more details.\n    (default: `[]`)\n  * `:sortable_fields` - a list of fields to build default sorters for. This option is simply a\n    shorthand syntax for using the `QueryElf.Plugins.AutomaticSorters` plugin with the given list\n    as the `fields` option. You should check the plugin's documentation for more details.\n    (default: `[]`)\n  * `:plugins` - a list of plugins that can be used to increment the query builder's\n    functionality. See `QueryElf.Plugin` for more details. (default: `[]`)\n\n#","ref":"QueryElf.html","title":"QueryElf","type":"behaviour"},{"doc":"defmodule MyQueryBuilder do\n      use QueryElf,\n        schema: MySchema,\n        searchable_fields: [:id, :name],\n        plugins: [\n          {QueryElf.Plugins.OffsetPagination, default_per_page: 10},\n          MyCustomPlugin\n        ]\n\n      def filter(:my_filter, value, _query) do\n        dynamic([s], s.some_field - ^value == 0)\n      end\n    end\n\n    MyQueryBuilder.build_query(id__in: [1, 2, 3], my_filter: 10)\n\n#","ref":"QueryElf.html#module-example","title":"Example - QueryElf","type":"behaviour"},{"doc":"Sometimes you have certain plugins that you wish to always use, while allowing some degree of\nfexibility to each individual query builder definition. For those scenarios, you can use something\nlike the following:\n\n    defmodule MyQueryElf do\n      defmacro __using__(opts) do\n        # Always define filters for the `id` field\n        searchable_fields = (opts[:searchable_fields] || []) ++ [:id]\n\n        # Always define a sorter for the `id` field\n        sortable_fields = (opts[:sortable_fields] || []) ++ [:id]\n\n        # Use a default per page of `20`, but allow the user to change this value\n        default_per_page = opts[:default_per_page] || 20\n\n        # Allow the user to include extra plugins\n        extra_plugins = opts[:plugins] || []\n\n        quote do\n          use QueryElf,\n            schema: unquote(opts[:schema]),\n            plugins: [\n              {QueryElf.Plugins.AutomaticFilters, fields: unquote(searchable_fields)},\n              {QueryElf.Plugins.AutomaticSorters, fields: unquote(sortable_fields)},\n              {QueryElf.Plugins.OffsetPagination, default_per_page: unquote(default_per_page)},\n              # put any other plugins here\n            ] ++ unquote(extra_plugins)\n        end\n      end\n    end\n\n    defmodule MyQueryBuilder do\n      use MyQueryElf,\n        schema: MySchema,\n        searchable_fields: ~w[id name age is_active roles]a\n    end\n\nUsing this strategy you can create a re-usable set of default plugins (and plugin configurations)\nthat best suits your application needs, while allowing you to use `QueryElf` without these\ndefaults if you ever need to.","ref":"QueryElf.html#module-sharing-plugins-and-configuration-accross-query-builders","title":"Sharing plugins and configuration accross query builders - QueryElf","type":"behaviour"},{"doc":"Should return the query builder metadata requested.","ref":"QueryElf.html#c:__query_builder__/1","title":"QueryElf.__query_builder__/1","type":"callback"},{"doc":"Should return the base query in this query builder. This base query will be used when defining the\n`empty_query/0` and `build_query/1` callbacks. If not defined, defaults to the supplied schema.\n\nThis is useful when dealing with logical deletion or other business rules that need to be followed\nevery time the query builder is used. Example:\n\n    defmodule UsersQueryBuilder do\n      use QueryElf,\n        schema: User,\n        searchable_fields: [:id, :name]\n\n      def base_query do\n        from u in User, where: is_nil(u.deleted_at)\n      end\n    end","ref":"QueryElf.html#c:base_query/0","title":"QueryElf.base_query/0","type":"callback"},{"doc":"Should receive a a keyword list or a map containing parameters and use it to build a query.","ref":"QueryElf.html#c:build_query/1","title":"QueryElf.build_query/1","type":"callback"},{"doc":"Same thing as `build_query/1`, but also receives some options for things like pagination and\nordering.","ref":"QueryElf.html#c:build_query/2","title":"QueryElf.build_query/2","type":"callback"},{"doc":"The same thing as `build_query/2`, but instead of building a new query it receives and extends\nan existing one.","ref":"QueryElf.html#c:build_query/3","title":"QueryElf.build_query/3","type":"callback"},{"doc":"Should return a query that when applied, returns nothing.","ref":"QueryElf.html#c:empty_query/0","title":"QueryElf.empty_query/0","type":"callback"},{"doc":"Receives an atom representing a filter, the parameter for the aforementioned filter and an Ecto\nquery. Should return an Ecto dynamic or a tuple containing an Ecto query and an Ecto dynamic.\n\nMost of the filtering should happen in the returned dynamic, and returning a query should only\nbe used when the query needs to be extended for the dynamic to make sense. Example:\n\n    # this is a simple comparison, so just returning a dynamic will suffice\n    def filter(:my_filter, value, _query) do\n      dynamic([s], s.some_field == ^value)\n    end\n\n    # this relies on a join, so the query must be extended accordingly\n    def filter(:my_other_filter, value, query) do\n      {\n        join(query, :left, [s], assoc(s, :some_relationship), as: :related),\n        dynamic([related: r], r.some_field == ^value)\n      }\n    end","ref":"QueryElf.html#c:filter/3","title":"QueryElf.filter/3","type":"callback"},{"doc":"Similar to `Ecto.Query.join/{4,5}`, but can be called multiple times with the same alias.\n\nNote that only the first join operation is performed, the subsequent ones that use the same alias\nare just ignored. Also note that because of this behaviour, its mandatory to specify an alias when\nusing this function.\n\nThis is helpful when you need to perform a join while building queries one filter at a time,\nbecause the same filter could be used multiple times or you could have multiple filters that\nrequire the same join.\n\nThis scenario poses a problem with how the `filter/3` callback work, as you need to return a\ndynamic with the filtering, which means that the join must have an alias, and by default Ecto\nraises an error when you add multiple joins with the same alias.\n\nTo solve this, it is recommended to use this macro instead of the default `Ecto.Query.join/{4,5}`.\nAs an added bonus, there will be only one join in the query that can be reused by multiple\nfilters.","ref":"QueryElf.html#reusable_join/5","title":"QueryElf.reusable_join/5","type":"macro"},{"doc":"Receives an atom representing a field to order by, the order direction, an extra argument to\nperform the ordering, and an Ecto query. Should return the Ecto query with the appropriate sorting\noptions applied.\n\nExample:\n\n      def sort(:name, direction, _arg, query) do\n        sort(query, [s], [{^direction, s.name}])\n      end","ref":"QueryElf.html#c:sort/4","title":"QueryElf.sort/4","type":"callback"},{"doc":"","ref":"QueryElf.html#t:filter/0","title":"QueryElf.filter/0","type":"type"},{"doc":"","ref":"QueryElf.html#t:metadata_type/0","title":"QueryElf.metadata_type/0","type":"type"},{"doc":"","ref":"QueryElf.html#t:options/0","title":"QueryElf.options/0","type":"type"},{"doc":"","ref":"QueryElf.html#t:sort_direction/0","title":"QueryElf.sort_direction/0","type":"type"},{"doc":"A behaviour to be followed when creating plugins that extends the query builders functionality.","ref":"QueryElf.Plugin.html","title":"QueryElf.Plugin","type":"behaviour"},{"doc":"Takes the query, the builder, and the options given to the `build_query` function build the query\nas arguments. Should return the modified query.","ref":"QueryElf.Plugin.html#c:build_query/3","title":"QueryElf.Plugin.build_query/3","type":"callback"},{"doc":"Called when QueryElf is used with the plugin.\n\nShould be used to inject required code in the builder module.","ref":"QueryElf.Plugin.html#c:using/1","title":"QueryElf.Plugin.using/1","type":"callback"},{"doc":"Plugin for automatically defining filters for a set of fields.\n\nIt accepts the following options:\n\n  - `:fields` - the list of fields for which to define filters. (required)\n\nThe defined filters will vary according to the field type in the schema:\n\n  * `id` or `binary_id`:\n    * `:$FIELD` - checks if the field is equal to the given value\n    * `:$FIELD__neq` - checks if the field is different from the given value\n    * `:$FIELD__in` - checks if the field is contained in the given enumerable\n    * `:$FIELD__not_in` - checks if the field is not contained in the given enumerable\n  * `boolean`:\n    * `:$FIELD` - checks if the field is equal to the given value\n  * `integer`, `float` or `decimal`:\n    * `:$FIELD` - checks if the field is equal to the given value\n    * `:$FIELD__neq` - checks if the field is different from the given value\n    * `:$FIELD__in` - checks if the field is contained in the given enumerable\n    * `:$FIELD__not_in` - checks if the field is not contained in the given enumerable\n    * `:$FIELD__gt` - checks if the field is greater than the given value\n    * `:$FIELD__lt` - checks if the field is lower than the given value\n    * `:$FIELD__gte` - checks if the field is greater than or equal to the given value\n    * `:$FIELD__lte` - checks if the field is lower than or equal to the given value\n  * `string`:\n    * `:$FIELD` - checks if the field is equal to the given value\n    * `:$FIELD__neq` - checks if the field is different from the given value\n    * `:$FIELD__in` - checks if the field is contained in the given enumerable\n    * `:$FIELD__not_in` - checks if the field is not contained in the given enumerable\n    * `:$FIELD__contains` - checks if the field contains the given string\n    * `:$FIELD__starts_with` - checks if the field starts with the given string\n    * `:$FIELD__ends_with` - checks if the field ends with the given string\n  * `{:array, inner_type}`:\n    * `:$FIELD` - checks if the field is equal to the given value\n    * `:$FIELD__neq` - checks if the field is different from the given value\n    * `:$FIELD__includes` - checks if the field includes the given value\n    * `:$FIELD__does_not_include` - checks if the field does not include the given value\n  * `date`, `time`, `naive_datetime`, `datetime`, `time_usec`, `naive_datetime_usec` or `datetime_usec`:\n    * `:$FIELD` - checks if the field is equal to the given value\n    * `:$FIELD__neq` - checks if the field is different from the given value\n    * `:$FIELD__in` - checks if the field is contained in the given enumerable\n    * `:$FIELD__not_in` - checks if the field is not contained in the given enumerable\n    * `:$FIELD__after` - checks if the field occurs after the given value\n    * `:$FIELD__before` - checks if the field occurs before the given value\n\nAny other types are simply ignored.\n\n#","ref":"QueryElf.Plugins.AutomaticFilters.html","title":"QueryElf.Plugins.AutomaticFilters","type":"module"},{"doc":"defmodule MyQueryBuilder do\n      use QueryElf,\n        schema: MySchema,\n        plugins: [\n          {QueryElf.Plugins.AutomaticFilters, fields: ~w[id name age is_active]a}\n        ]\n    end\n\n    MyQueryBuilder.build_query(id__in: [1, 2, 3], name__starts_with: \"John\")","ref":"QueryElf.Plugins.AutomaticFilters.html#module-example","title":"Example - QueryElf.Plugins.AutomaticFilters","type":"module"},{"doc":"Plugin for automatically defining sorters for a set of fields.\n\nIt accepts the following options:\n\n  - `:fields` - the list of fields for which to define sorters. (required)\n\n#","ref":"QueryElf.Plugins.AutomaticSorters.html","title":"QueryElf.Plugins.AutomaticSorters","type":"module"},{"doc":"This definition:\n\n    defmodule MyQueryBuilder do\n      use QueryElf,\n        schema: MySchema,\n        plugins: [\n          {QueryElf.Plugins.AutomaticSorters, fields: ~w[inserted_at]a}\n        ]\n    end\n\nis equivalent to:\n\n    defmodule MyQueryBuilder do\n      use QueryElf,\n        schema: MySchema\n\n      def sort(:inserted_at, direction, _args, query) do\n        case direction do\n          :asc -> order_by(query, asc: :inserted_at)\n          :desc -> order_by(query, desc: :inserted_at)\n        end\n      end\n    end","ref":"QueryElf.Plugins.AutomaticSorters.html#module-example","title":"Example - QueryElf.Plugins.AutomaticSorters","type":"module"},{"doc":"Plugin to enable limit-offset based pagination in any query builder.\n\nIt accepts the following options:\n\n  * `:default_per_page` - the default value of items per page (default: `25`)\n\n#","ref":"QueryElf.Plugins.OffsetPagination.html","title":"QueryElf.Plugins.OffsetPagination","type":"module"},{"doc":"defmodule MySchemaQueryBuilder do\n      use QueryElf,\n        schema: MySchema,\n        plugins: [QueryElf.Plugins.OffsetPagination]\n    end\n\n    MySchemaQueryBuilder.build_query([], page: 2, per_page: 25)","ref":"QueryElf.Plugins.OffsetPagination.html#module-example","title":"Example - QueryElf.Plugins.OffsetPagination","type":"module"},{"doc":"# Basic Usage\n\nSuppose you have a `users` schema that is defined as below:\n\n```elixir\ndefmodule Accounts.User do\n  schema \"users\" do\n    field :email, :string\n    field :password_hash, :string\n\n    field :name, :string\n    field :is_active, :boolean\n\n    field :signed_up_at, :naive_datetime\n  end\nend\n```","ref":"basic_usage.html","title":"Basic Usage","type":"extras"},{"doc":"Suppose you wish to enable filtering by `email`, `name`, `is_active`, and `signed_up_at`. Then you can define a `QueryBuilder` module as below:\n\n```elixir\ndefmodule Accounts.User.QueryBuilder do\n  use QueryElf,\n    schema: Accounts.User,\n    searchable_fields: [:email, :name, :is_active, :signed_up_at]\nend\n```\n\nThe `schema` specifies the ecto schema for which the query builder is being created. The `searchable_fields` specifies fields that can be filtered on. With the above definition, you can now use the query builder's capabailities as below:\n\n```elixir\nUser.QueryBuilder.build_query(email: \"a@b.com\")\n|> Repo.one\n\nUser.QueryBuilder.build_query(name__in: [\"Raphael Costa\", \"Jose Valim\"])\n|> Repo.all()\n\nUser.QueryBuilder.build_query(is_active: true)\n|> Repo.all()\n\nUser.QueryBuilder.build_query(signed_up_at__before: ~N[2000-01-01 23:00:07])\n|> Repo.all()\n```\n\nYou can also combine multiple conditions/filters together:\n\n```elixir\n# Conditions are `anded` by default\nUser.QueryBuilder.build_query(name__in: [\"Raphael Costa\", \"Jose Valim\"], is_active: true)\n|> Repo.all()\n\n# But you can change to using `or`\nUser.QueryBuilder.build_query(_or: [name: \"Mr A\", email: \"a@b.com\"])\n|> Repo.all()\n\n# Or combine them together to create a complex query\nUser.QueryBuilder.build_query(\n  _or: [\n    name: \"Mr A\",\n    _and: [\n      email__in: [\"a@b.com\", \"a@c.com\"],\n      is_active: true\n    ]\n  ]\n)\n|> Repo.all()\n```\n\nFor more information about what filters (`__in`, `__before`, etc) are defined for each field type, check the documentation for the `searchable_fields` option [here](`QueryElf`).","ref":"basic_usage.html#basic-filtering","title":"Basic filtering - Basic Usage","type":"extras"},{"doc":"Assuming you have the same user schema defined above, suppose you wish to allow sorting users by `name`, `email`, and `signed_up_at`. QueryElf can be used to achieve this as below:\n\n```elixir\ndefmodule Accounts.User.QueryBuilder do\n  use QueryElf,\n    schema: Accounts.User,\n    searchable_fields: ~w[email name is_active signed_up_at]a,\n    sortable_fields: ~w[email name signed_up_at]a\nend\n```\n\nThen sorting capabilities can be included in `build_query` as below:\n\n```elixir\nAccounts.User.QueryBuilder.build_query([],\n  order: [%{field: :name, direction: :asc}, %{field: :signed_up_at, direction: :desc}]\n)\n|> Repo.all()\n\n\n# Shorthand expression for ordering example above\nAccounts.User.QueryBuilder.build_query([],\n  order: [asc: :my_int, desc: :my_bool]\n)\n|> Repo.all()\n```","ref":"basic_usage.html#basic-ordering-sorting","title":"Basic ordering/sorting - Basic Usage","type":"extras"},{"doc":"# Advanced Usage\n\nSuppose you have a `users` schema that is defined as below:\n\n```elixir\ndefmodule Accounts.User do\n  schema \"users\" do\n    field :email, :string\n    field :password_hash, :string\n\n    field :name, :string\n    field :is_active, :boolean\n\n    field :signed_up_at, :naive_datetime\n    field :last_logged_in_at, :naive_datetime\n  end\nend\n```\n\nThen a basic filtering and ordering query builder can be defined as:\n\n```elixir\ndefmodule Accounts.User.QueryBuilder do\n  use QueryElf,\n    schema: Accounts.User,\n    searchable_fields: ~w[email name is_active signed_up_at]a,\n    sortable_fields: ~w[email name signed_up_at]a\nend\n```\n\nWhere filtering is possible on `email`, `name`, `is_active`, and `signed_up_at`, and sorting is possible on `email`, `name`, and `signed_up_at`. If this is confusing for you, it would be a good idea to read through the basic usage documentation.","ref":"advanced_usage.html","title":"Advanced Usage","type":"extras"},{"doc":"Suppose you wish to implement a custom filter to identify recently active users defined as users that logged in in the last `n` days. You can define and use such a filter as below:\n\n```elixir\ndefmodule Accounts.User.QueryBuilder do\n  use QueryElf,\n    schema: Accounts.User,\n    searchable_fields: ~w[email name is_active signed_up_at]a,\n    sortable_fields: ~w[email name signed_up_at]a\n\n  def filter(:active_in_last_days, days, _query) do\n    earliest_logged_in_at =\n      NaiveDateTime.add(\n        NaiveDateTime.utc_now(),\n        days * 24 * 3600 * -1,\n        :second\n      )\n\n    dynamic([u], u.last_logged_in_at > ^earliest_logged_in_at)\n  end\nend\n\nAccounts.User.QueryBuilder.build_query(active_in_last_days: 30)\n|> Repo.all()\n```\n\nSuppose you also have a notification preferences schema and a foriegn key from users schema to the preferences schema and you want to implement custom filters to identify recently active users that are happy to receive newsletters, then that can be achieved as below:\n\n```elixir\ndefmodule Accounts.User do\n  schema \"users\" do\n    field :email, :string\n    field :password_hash, :string\n\n    field :name, :string\n    field :is_active, :boolean\n\n    field :signed_up_at, :naive_datetime\n    field :last_logged_in_at, :naive_datetime\n\n    has_one :notification_preferences, NotificationPreferences\n  end\nend\n\ndefmodule Accounts.NotificationPreferences do\n  schema \"notification_preferences\" do\n    field :send_product_updates, :boolean\n    field :send_newsletters, :boolean\n  end\nend\n\ndefmodule Accounts.User.QueryBuilder do\n  use QueryElf,\n    schema: Accounts.User,\n    searchable_fields: ~w[email name is_active signed_up_at]a,\n    sortable_fields: ~w[email name signed_up_at]a\n\n  def filter(:active_in_last_days, days, _query) do\n    earliest_logged_in_at =\n      NaiveDateTime.add(\n        NaiveDateTime.utc_now(),\n        days * 24 * 3600 * -1,\n        :second\n      )\n\n    dynamic([u], u.last_logged_in_at > ^earliest_logged_in_at)\n  end\n\n  def filter(:accepts_newsletters, value, query) do\n    {\n      # Use reusable_join instead of join as explained in the next section\n      join(query, assoc(r, :notification_preferences), as: :preferences),\n      dynamic([preferences: p], p.send_newsletters == ^value)\n    }\n  end\nend\n\nAccounts.User.QueryBuilder.build_query(active_in_last_days: 30, accepts_newsletters: true)\n|> Repo.all()\n```","ref":"advanced_usage.html#defining-custom-filters","title":"Defining custom filters - Advanced Usage","type":"extras"},{"doc":"Given the module definition of `User.QueryBuilder` above, if you wrote the following query:\n\n```elixir\nAccounts.User.QueryBuilder.build_query(\n  _or: [\n    accepts_newsletters: true,\n    _and: [active_in_last_days: 5, accepts_newsletters: false]\n  ]\n)\n```\n\nThen ecto with throw an error as it would result in two named joins having the same name. This is because both `accepts_newsletters: true` and `accepts_newsletters: false` add a join each between `users` and `notification_preferences` with the name `preferences` and ecto doesn't allow two joins with the same name.\n\nWhat you really want is that a single join is added to the query as they both are essentially the same join. This is achieved by the `reusable_join` macro.\n\n```elixir\nimport QueryElf\n\ndefmodule Accounts.User.QueryBuilder do\n  use QueryElf,\n    schema: Accounts.User,\n    searchable_fields: ~w[email name is_active signed_up_at]a,\n    sortable_fields: ~w[email name signed_up_at]a\n\n  def filter(:accepts_newsletters, value, query) do\n    {\n      reusable_join(query, assoc(r, :notification_preferences), as: :preferences),\n      dynamic([preferences: p], p.send_newsletters == ^value)\n    }\n  end\nend\n```\n\nInternally, the `reusable_join` macro checks if a join with the same name has already been added to the query and if so it skips adding the join. Thus, if you have two custom filters that should be using the same join, you can ensure that happens by giving the joins in the filters the same name.\n\nFor example, suppose you want to implement custom filters to allow fetching users that are happy to receive newsletters and product updates. In this case, you would need to build two different filters so that they can be used independently and also be combined. But you want to ensure that when combined, the filters use a single join instead of multiple joins on `users` and `notification_preferences`. This is achieved by using `reusable_join` with the same join name in both the custom filters.\n\n```elixir\ndefmodule Accounts.User.QueryBuilder do\n  use QueryElf,\n    schema: Accounts.User,\n    searchable_fields: ~w[email name is_active signed_up_at]a,\n    sortable_fields: ~w[email name signed_up_at]a\n\n  def filter(:accepts_newsletters, value, query) do\n    {\n      reusable_join(query, assoc(r, :notification_preferences), as: :preferences),\n      dynamic([preferences: p], p.send_newsletters == ^value)\n    }\n  end\n\n  def filter(:accepts_product_updates, value, query) do\n    {\n      reusable_join(query, assoc(r, :notification_preferences), as: :preferences),\n      dynamic([preferences: p], p.send_product_updates == ^value)\n    }\n  end\nend\n\nAccounts.User.QueryBuilder.build_query(accepts_newsletters: true)\n|> Repo.all()\n\nAccounts.User.QueryBuilder.build_query(accepts_product_updates: true)\n|> Repo.all()\n\nAccounts.User.QueryBuilder.build_query(accepts_newsletters: true, accepts_product_updates: true)\n|> Repo.all()\n```","ref":"advanced_usage.html#using-reusable-joins","title":"Using reusable joins - Advanced Usage","type":"extras"}]}